@@ -15,10 +15,19 @@ to_from_bytes!(PlutusScript);
 
 #[wasm_bindgen]
 impl PlutusScript {
+    /**
+     * Creates a new Plutus script from the RAW bytes of the compiled script.
+     * This does NOT include any CBOR encoding around these bytes (e.g. from "cborBytes" in cardano-cli)
+     * If you creating this from those you should use PlutusScript::from_bytes() instead.
+     */
     pub fn new(bytes: Vec<u8>) -> PlutusScript {
         Self(bytes)
     }
 
+    /**
+     * The raw bytes of this compiled Plutus script.
+     * If you need "cborBytes" for cardano-cli use PlutusScript::to_bytes() instead.
+     */
     pub fn bytes(&self) -> Vec<u8> {
         self.0.clone()
     }
@@ -52,7 +61,7 @@ impl PlutusScripts {
 #[wasm_bindgen]
 #[derive(Clone, Debug, Eq, Ord, PartialEq, PartialOrd)]
 pub struct ConstrPlutusData {
-    tag: Int,
+    alternative: BigNum,
     data: PlutusList,
 }
 
@@ -60,28 +69,52 @@ to_from_bytes!(ConstrPlutusData);
 
 #[wasm_bindgen]
 impl ConstrPlutusData {
-    pub fn tag(&self) -> Int {
-        self.tag.clone()
+    pub fn alternative(&self) -> BigNum {
+        self.alternative.clone()
     }
 
     pub fn data(&self) -> PlutusList {
         self.data.clone()
     }
 
-    pub fn new(tag: Int, data: &PlutusList) -> Self {
+    pub fn new(alternative: &BigNum, data: &PlutusList) -> Self {
         Self {
-            tag,
+            alternative: alternative.clone(),
             data: data.clone(),
         }
     }
 }
 
 impl ConstrPlutusData {
-    fn is_tag_compact(tag: i128) -> bool {
-        (tag >= 121 && tag <= 127) || (tag >= 1280 && tag <= 1400)
+    // see: https://github.com/input-output-hk/plutus/blob/1f31e640e8a258185db01fa899da63f9018c0e85/plutus-core/plutus-core/src/PlutusCore/Data.hs#L61
+    // We don't directly serialize the alternative in the tag, instead the scheme is:
+    // - Alternatives 0-6 -> tags 121-127, followed by the arguments in a list
+    // - Alternatives 7-127 -> tags 1280-1400, followed by the arguments in a list
+    // - Any alternatives, including those that don't fit in the above -> tag 102 followed by a list containing
+    //   an unsigned integer for the actual alternative, and then the arguments in a (nested!) list.
+    const GENERAL_FORM_TAG: u64 = 102;
+
+    // None -> needs general tag serialization, not compact
+    fn alternative_to_compact_cbor_tag(alt: u64) -> Option<u64> {
+        if alt <= 6 {
+            Some(121 + alt)
+        } else if alt >= 7 && alt <= 127 {
+            Some(1280 - 7 + alt)
+        } else {
+            None
+        }
     }
 
-    const GENERAL_FORM_TAG: u64 = 102;
+    // None -> General tag(=102) OR Invalid CBOR tag for this scheme
+    fn compact_cbor_tag_to_alternative(cbor_tag: u64) -> Option<u64> {
+        if cbor_tag >= 121 && cbor_tag <= 127 {
+            Some(cbor_tag - 121)
+        } else if cbor_tag >= 1280 && cbor_tag <= 1400 {
+            Some(cbor_tag - 1280 + 7)
+        } else {
+            None
+        }
+    }
 }
 
 const COST_MODEL_OP_COUNT: usize = 166;
@@ -146,6 +179,33 @@ impl Costmdls {
     pub fn keys(&self) -> Languages {
         Languages(self.0.iter().map(|(k, _v)| k.clone()).collect::<Vec<_>>())
     }
+
+    pub(crate) fn language_views_encoding(&self) -> Vec<u8> {
+        let mut serializer = Serializer::new_vec();
+        let mut keys_bytes: Vec<(Language, Vec<u8>)> = self.0.iter().map(|(k, _v)| (k.clone(), k.to_bytes())).collect();
+        // keys must be in canonical ordering first
+        keys_bytes.sort_by(|lhs, rhs| match lhs.1.len().cmp(&rhs.1.len()) {
+            std::cmp::Ordering::Equal => lhs.1.cmp(&rhs.1),
+            len_order => len_order,
+        });
+        serializer.write_map(cbor_event::Len::Len(self.0.len() as u64)).unwrap();
+        for (key, key_bytes) in keys_bytes.iter() {
+            serializer.write_bytes(key_bytes).unwrap();
+            let cost_model = self.0.get(&key).unwrap();
+            // Due to a bug in the cardano-node input-output-hk/cardano-ledger-specs/issues/2512
+            // we must use indefinite length serialization in this inner bytestring to match it
+            let mut cost_model_serializer = Serializer::new_vec();
+            cost_model_serializer.write_array(cbor_event::Len::Indefinite).unwrap();
+            for cost in &cost_model.0 {
+                cost.serialize(&mut cost_model_serializer).unwrap();
+            }
+            cost_model_serializer.write_special(cbor_event::Special::Break).unwrap();
+            serializer.write_bytes(cost_model_serializer.finalize()).unwrap();
+        }
+        let out = serializer.finalize();
+        println!("language_views = {}", hex::encode(out.clone()));
+        out
+    }
 }
 
 #[wasm_bindgen]
@@ -274,7 +334,10 @@ impl PlutusMap {
     }
 
     pub fn keys(&self) -> PlutusList {
-        PlutusList(self.0.iter().map(|(k, _v)| k.clone()).collect::<Vec<_>>())
+        PlutusList {
+            elems: self.0.iter().map(|(k, _v)| k.clone()).collect::<Vec<_>>(),
+            definite_encoding: None,
+        }
     }
 }
 
@@ -299,34 +362,54 @@ enum PlutusDataEnum {
 
 #[wasm_bindgen]
 #[derive(Clone, Debug, Eq, Ord, PartialEq, PartialOrd)]
-pub struct PlutusData(PlutusDataEnum);
+pub struct PlutusData {
+    datum: PlutusDataEnum,
+    // We should always preserve the original datums when deserialized as this is NOT canonicized
+    // before computing datum hashes. So this field stores the original bytes to re-use.
+    original_bytes: Option<Vec<u8>>,
+}
 
 to_from_bytes!(PlutusData);
 
 #[wasm_bindgen]
 impl PlutusData {
     pub fn new_constr_plutus_data(constr_plutus_data: &ConstrPlutusData) -> Self {
-        Self(PlutusDataEnum::ConstrPlutusData(constr_plutus_data.clone()))
+        Self {
+            datum: PlutusDataEnum::ConstrPlutusData(constr_plutus_data.clone()),
+            original_bytes: None,
+        }
     }
 
     pub fn new_map(map: &PlutusMap) -> Self {
-        Self(PlutusDataEnum::Map(map.clone()))
+        Self {
+            datum: PlutusDataEnum::Map(map.clone()),
+            original_bytes: None,
+        }
     }
 
     pub fn new_list(list: &PlutusList) -> Self {
-        Self(PlutusDataEnum::List(list.clone()))
+        Self {
+            datum: PlutusDataEnum::List(list.clone()),
+            original_bytes: None,
+        }
     }
 
     pub fn new_integer(integer: &BigInt) -> Self {
-        Self(PlutusDataEnum::Integer(integer.clone()))
+        Self {
+            datum: PlutusDataEnum::Integer(integer.clone()),
+            original_bytes: None,
+        }
     }
 
     pub fn new_bytes(bytes: Vec<u8>) -> Self {
-        Self(PlutusDataEnum::Bytes(bytes))
+        Self {
+            datum: PlutusDataEnum::Bytes(bytes),
+            original_bytes: None,
+        }
     }
 
     pub fn kind(&self) -> PlutusDataKind {
-        match &self.0 {
+        match &self.datum {
             PlutusDataEnum::ConstrPlutusData(_) => PlutusDataKind::ConstrPlutusData,
             PlutusDataEnum::Map(_) => PlutusDataKind::Map,
             PlutusDataEnum::List(_) => PlutusDataKind::List,
@@ -336,35 +419,35 @@ impl PlutusData {
     }
 
     pub fn as_constr_plutus_data(&self) -> Option<ConstrPlutusData> {
-        match &self.0 {
+        match &self.datum {
             PlutusDataEnum::ConstrPlutusData(x) => Some(x.clone()),
             _ => None,
         }
     }
 
     pub fn as_map(&self) -> Option<PlutusMap> {
-        match &self.0 {
+        match &self.datum {
             PlutusDataEnum::Map(x) => Some(x.clone()),
             _ => None,
         }
     }
 
     pub fn as_list(&self) -> Option<PlutusList> {
-        match &self.0 {
+        match &self.datum {
             PlutusDataEnum::List(x) => Some(x.clone()),
             _ => None,
         }
     }
 
     pub fn as_integer(&self) -> Option<BigInt> {
-        match &self.0 {
+        match &self.datum {
             PlutusDataEnum::Integer(x) => Some(x.clone()),
             _ => None,
         }
     }
 
     pub fn as_bytes(&self) -> Option<Vec<u8>> {
-        match &self.0 {
+        match &self.datum {
             PlutusDataEnum::Bytes(x) => Some(x.clone()),
             _ => None,
         }
@@ -373,26 +456,36 @@ impl PlutusData {
 
 #[wasm_bindgen]
 #[derive(Clone, Debug, Eq, Ord, PartialEq, PartialOrd)]
-pub struct PlutusList(Vec<PlutusData>);
+pub struct PlutusList {
+    elems: Vec<PlutusData>,
+    // We should always preserve the original datums when deserialized as this is NOT canonicized
+    // before computing datum hashes. This field will default to cardano-cli behavior if None
+    // and will re-use the provided one if deserialized, unless the list is modified.
+    definite_encoding: Option<bool>,
+}
 
 to_from_bytes!(PlutusList);
 
 #[wasm_bindgen]
 impl PlutusList {
     pub fn new() -> Self {
-        Self(Vec::new())
+        Self {
+            elems: Vec::new(),
+            definite_encoding: None,
+        }
     }
 
     pub fn len(&self) -> usize {
-        self.0.len()
+        self.elems.len()
     }
 
     pub fn get(&self, index: usize) -> PlutusData {
-        self.0[index].clone()
+        self.elems[index].clone()
     }
 
     pub fn add(&mut self, elem: &PlutusData) {
-        self.0.push(elem.clone());
+        self.elems.push(elem.clone());
+        self.definite_encoding = None;
     }
 }
 
@@ -579,15 +672,15 @@ impl Deserialize for PlutusScripts {
 // TODO: write tests for this hand-coded implementation?
 impl cbor_event::se::Serialize for ConstrPlutusData {
     fn serialize<'se, W: Write>(&self, serializer: &'se mut Serializer<W>) -> cbor_event::Result<&'se mut Serializer<W>> {
-        if Self::is_tag_compact(self.tag.0) {
+        if let Some(compact_tag) = Self::alternative_to_compact_cbor_tag(from_bignum(&self.alternative)) {
             // compact form
-            serializer.write_tag(self.tag.0 as u64)?;
+            serializer.write_tag(compact_tag as u64)?;
             self.data.serialize(serializer)
         } else {
             // general form
             serializer.write_tag(Self::GENERAL_FORM_TAG)?;
             serializer.write_array(cbor_event::Len::Len(2))?;
-            self.tag.serialize(serializer)?;
+            self.alternative.serialize(serializer)?;
             self.data.serialize(serializer)
         }
     }
@@ -596,13 +689,13 @@ impl cbor_event::se::Serialize for ConstrPlutusData {
 impl Deserialize for ConstrPlutusData {
     fn deserialize<R: BufRead + Seek>(raw: &mut Deserializer<R>) -> Result<Self, DeserializeError> {
         (|| -> Result<_, DeserializeError> {
-            let (tag, data) = match raw.tag()? {
+            let (alternative, data) = match raw.tag()? {
                 // general form
                 Self::GENERAL_FORM_TAG => {
                     let len = raw.array()?;
                     let mut read_len = CBORReadLen::new(len);
                     read_len.read_elems(2)?;
-                    let tag = Int::deserialize(raw)?;
+                    let alternative = BigNum::deserialize(raw)?;
                     let data = (|| -> Result<_, DeserializeError> {
                         Ok(PlutusList::deserialize(raw)?)
                     })().map_err(|e| e.annotate("datas"))?;
@@ -613,17 +706,22 @@ impl Deserialize for ConstrPlutusData {
                             _ => return Err(DeserializeFailure::EndingBreakMissing.into()),
                         },
                     }
-                    (tag, data)
+                    (alternative, data)
                 },
                 // concise form
-                tag if Self::is_tag_compact(tag.into()) => (Int::new(&to_bignum(tag)), PlutusList::deserialize(raw)?),
-                invalid_tag => return Err(DeserializeFailure::TagMismatch{
-                    found: invalid_tag,
-                    expected: Self::GENERAL_FORM_TAG,
-                }.into()),
+                tag => {
+                    if let Some(alternative) = Self::compact_cbor_tag_to_alternative(tag) {
+                        (to_bignum(alternative), PlutusList::deserialize(raw)?)
+                    } else {
+                        return Err(DeserializeFailure::TagMismatch{
+                            found: tag,
+                            expected: Self::GENERAL_FORM_TAG,
+                        }.into());
+                    }
+                },
             };
             Ok(ConstrPlutusData{
-                tag,
+                alternative,
                 data,
             })
         })().map_err(|e| e.annotate("ConstrPlutusData"))
@@ -920,22 +1018,48 @@ impl Deserialize for PlutusDataEnum {
 
 impl cbor_event::se::Serialize for PlutusData {
     fn serialize<'se, W: Write>(&self, serializer: &'se mut Serializer<W>) -> cbor_event::Result<&'se mut Serializer<W>> {
-        self.0.serialize(serializer)
+        match &self.original_bytes {
+            Some(bytes) => serializer.write_raw_bytes(bytes),
+            None => self.datum.serialize(serializer),
+        }
     }
 }
 
 impl Deserialize for PlutusData {
     fn deserialize<R: BufRead + Seek>(raw: &mut Deserializer<R>) -> Result<Self, DeserializeError> {
-        Ok(Self(PlutusDataEnum::deserialize(raw)?))
+        // these unwraps are fine since we're seeking the current position
+        let before = raw.as_mut_ref().seek(SeekFrom::Current(0)).unwrap();
+        let datum = PlutusDataEnum::deserialize(raw)?;
+        let after = raw.as_mut_ref().seek(SeekFrom::Current(0)).unwrap();
+        let bytes_read = (after - before) as usize;
+        raw.as_mut_ref().seek(SeekFrom::Start(before)).unwrap();
+        // these unwraps are fine since we read the above already
+        let original_bytes = raw.as_mut_ref().fill_buf().unwrap()[..bytes_read].to_vec();
+        raw.as_mut_ref().consume(bytes_read);
+        Ok(Self {
+            datum,
+            original_bytes: Some(original_bytes),
+        })
     }
 }
 
 impl cbor_event::se::Serialize for PlutusList {
     fn serialize<'se, W: Write>(&self, serializer: &'se mut Serializer<W>) -> cbor_event::Result<&'se mut Serializer<W>> {
-        serializer.write_array(cbor_event::Len::Len(self.0.len() as u64))?;
-        for element in &self.0 {
+        let use_definite_encoding = match self.definite_encoding {
+            Some(definite) => definite,
+            None => self.elems.is_empty(),
+        };
+        if use_definite_encoding {
+            serializer.write_array(cbor_event::Len::Len(self.elems.len() as u64))?;
+        } else {
+            serializer.write_array(cbor_event::Len::Indefinite)?;
+        }
+        for element in &self.elems {
             element.serialize(serializer)?;
         }
+        if !use_definite_encoding {
+            serializer.write_special(cbor_event::Special::Break)?;
+        }
         Ok(serializer)
     }
 }
@@ -943,7 +1067,7 @@ impl cbor_event::se::Serialize for PlutusList {
 impl Deserialize for PlutusList {
     fn deserialize<R: BufRead + Seek>(raw: &mut Deserializer<R>) -> Result<Self, DeserializeError> {
         let mut arr = Vec::new();
-        (|| -> Result<_, DeserializeError> {
+        let len = (|| -> Result<_, DeserializeError> {
             let len = raw.array()?;
             while match len { cbor_event::Len::Len(n) => arr.len() < n as usize, cbor_event::Len::Indefinite => true, } {
                 if raw.cbor_type()? == CBORType::Special {
@@ -952,9 +1076,12 @@ impl Deserialize for PlutusList {
                 }
                 arr.push(PlutusData::deserialize(raw)?);
             }
-            Ok(())
+            Ok(len)
         })().map_err(|e| e.annotate("PlutusList"))?;
-        Ok(Self(arr))
+        Ok(Self {
+            elems: arr,
+            definite_encoding: Some(len != cbor_event::Len::Indefinite),
+        })
     }
 }
 
@@ -1104,3 +1231,60 @@ impl Deserialize for Strings {
         Ok(Self(arr))
     }
 }
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+    use hex::*;
+
+    #[test]
+    pub fn plutus_constr_data() {
+        let constr_0 = PlutusData::new_constr_plutus_data(
+            &ConstrPlutusData::new(&to_bignum(0), &PlutusList::new())
+        );
+        let constr_0_hash = hex::encode(hash_plutus_data(&constr_0).to_bytes());
+        assert_eq!(constr_0_hash, "923918e403bf43c34b4ef6b48eb2ee04babed17320d8d1b9ff9ad086e86f44ec");
+        let constr_0_roundtrip = PlutusData::from_bytes(constr_0.to_bytes()).unwrap();
+        // TODO: do we want semantic equality or bytewise equality?
+        //assert_eq!(constr_0, constr_0_roundtrip);
+        let constr_1854 = PlutusData::new_constr_plutus_data(
+            &ConstrPlutusData::new(&to_bignum(1854), &PlutusList::new())
+        );
+        let constr_1854_roundtrip = PlutusData::from_bytes(constr_1854.to_bytes()).unwrap();
+        //assert_eq!(constr_1854, constr_1854_roundtrip);
+    }
+
+    #[test]
+    pub fn plutus_list_serialization_cli_compatibility() {
+        // mimic cardano-cli array encoding, see https://github.com/Emurgo/cardano-serialization-lib/issues/227
+        let datum_cli = "d8799f4100d8799fd8799fd8799f581cffffffffffffffffffffffffffffffffffffffffffffffffffffffffffd8799fd8799fd8799f581cffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffd87a80ff1a002625a0d8799fd879801a000f4240d87a80ffff";
+        let datum = PlutusData::from_bytes(Vec::from_hex(datum_cli).unwrap()).unwrap();
+        assert_eq!(datum_cli, hex::encode(datum.to_bytes()));
+
+        // encode empty arrays as fixed
+        assert_eq!("80", hex::encode(PlutusList::new().to_bytes()));
+
+        // encode arrays as indefinite length array
+        let mut list = PlutusList::new();
+        list.add(&PlutusData::new_integer(&BigInt::from_str("1").unwrap()));
+        assert_eq!("9f01ff", hex::encode(list.to_bytes()));
+
+        // witness_set should have fixed length array
+        let mut witness_set = TransactionWitnessSet::new();
+        witness_set.set_plutus_data(&list);
+        assert_eq!("a1048101", hex::encode(witness_set.to_bytes()));
+
+        list = PlutusList::new();
+        list.add(&datum);
+        witness_set.set_plutus_data(&list);
+        assert_eq!(format!("a10481{}", datum_cli), hex::encode(witness_set.to_bytes()));
+    }
+
+    #[test]
+    pub fn plutus_datums_respect_deserialized_encoding() {
+        let orig_bytes = Vec::from_hex("81d8799f581ce1cbb80db89e292269aeb93ec15eb963dda5176b66949fe1c2a6a38da140a1401864ff").unwrap();
+        let datums = PlutusList::from_bytes(orig_bytes.clone()).unwrap();
+        let new_bytes = datums.to_bytes();
+        assert_eq!(orig_bytes, new_bytes);
+    }
+}
